#include "Common.h"

#include "DODVoxels.h"
#include "Framework.h"

namespace Exp
{

	void DODVoxels::Init()
	{
		m_VoxelSize = 20.f;

		m_VoxelDataSize = Framework::s_VoxelTotal * 4 * 6; // 4 points per quad, 6 quads per voxel

		m_VoxelData = new VoxelData;
		m_VoxelData->x = new float[Framework::s_VoxelTotal];
		m_VoxelData->y = new float[Framework::s_VoxelTotal];
		m_VoxelData->z = new float[Framework::s_VoxelTotal];
		m_VoxelData->r = new float[Framework::s_VoxelTotal];
		m_VoxelData->g = new float[Framework::s_VoxelTotal];
		m_VoxelData->b = new float[Framework::s_VoxelTotal];

		m_VoxelRenderData = new VoxelData;
		m_VoxelRenderData->x = new float[Framework::s_VoxelTotal];
		m_VoxelRenderData->y = new float[Framework::s_VoxelTotal];
		m_VoxelRenderData->z = new float[Framework::s_VoxelTotal];
		m_VoxelRenderData->r = new float[Framework::s_VoxelTotal];
		m_VoxelRenderData->g = new float[Framework::s_VoxelTotal];
		m_VoxelRenderData->b = new float[Framework::s_VoxelTotal];

		m_VoxelRayCastData = new VoxelData;
		m_VoxelRayCastData->x = new float[Framework::s_VoxelTotal];
		m_VoxelRayCastData->y = new float[Framework::s_VoxelTotal];
		m_VoxelRayCastData->z = new float[Framework::s_VoxelTotal];
		m_VoxelRayCastData->r = new float[Framework::s_VoxelTotal];
		m_VoxelRayCastData->g = new float[Framework::s_VoxelTotal];
		m_VoxelRayCastData->b = new float[Framework::s_VoxelTotal];

		m_VoxelRayCast = new RayCastData;
		m_VoxelRayCast->ox = new float[Framework::s_VoxelTotal];
		m_VoxelRayCast->oy = new float[Framework::s_VoxelTotal];
		m_VoxelRayCast->oz = new float[Framework::s_VoxelTotal];
		m_VoxelRayCast->dx = new float[Framework::s_VoxelTotal];
		m_VoxelRayCast->dy = new float[Framework::s_VoxelTotal];
		m_VoxelRayCast->dz = new float[Framework::s_VoxelTotal];

		m_VoxelDebugRayCast = new Vec3[Framework::s_VoxelTotal * 2];

		// generate voxels

		GenerateVoxels(m_VoxelData, Framework::s_VoxelTotal);

		// offsets

		float halfsize = m_VoxelSize / 2.f;

		m_VoxelOffset[0] = Vec3(-halfsize, -halfsize,  halfsize);
		m_VoxelOffset[1] = Vec3( halfsize, -halfsize,  halfsize);
		m_VoxelOffset[2] = Vec3(-halfsize,  halfsize,  halfsize);
		m_VoxelOffset[3] = Vec3( halfsize,  halfsize,  halfsize);

		m_VoxelOffset[4] = Vec3(-halfsize, -halfsize, -halfsize);
		m_VoxelOffset[5] = Vec3( halfsize, -halfsize, -halfsize);
		m_VoxelOffset[6] = Vec3(-halfsize,  halfsize, -halfsize);
		m_VoxelOffset[7] = Vec3( halfsize,  halfsize, -halfsize);

		// generate initial buffer

		m_VoxelRenderTotal = 0;

		// voxel buffer

		glGenBuffers(1, &m_VoxelVBOVertex); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex); CGLE();
		glBufferData(GL_ARRAY_BUFFER, m_VoxelDataSize * sizeof(Vec3), NULL, GL_STREAM_DRAW); CGLE();

		glGenBuffers(1, &m_VoxelVBOColor); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor); CGLE();
		glBufferData(GL_ARRAY_BUFFER, m_VoxelDataSize * sizeof(Vec3), NULL, GL_STREAM_DRAW); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0); CGLE();
	}

	void DODVoxels::Tick(float a_DT)
	{
		
	}

	void DODVoxels::Reload()
	{
		GenerateVoxels(m_VoxelData, Framework::s_VoxelTotal);
	}

	void DODVoxels::PostTick()
	{
		if (Framework::s_RayCast)
		{
			Vec3 target(
				Math::CosDeg(Framework::s_TargetRotation) * 500.f, 
				10.f, 
				Math::SinDeg(Framework::s_TargetRotation) * 500.f
			);

			GenerateRayCast(
				m_VoxelRayCast, m_VoxelRayCastTotal,
				m_VoxelData, Framework::s_VoxelCurrentTotal, 
				target
			);

			SolveRayCast(
				m_VoxelRenderData, m_VoxelRenderTotal,
				m_VoxelRayCast, m_VoxelRayCastTotal,
				m_VoxelData, Framework::s_VoxelCurrentTotal
			);

			m_VoxelRenderDataUsed = m_VoxelRenderData;
		}
		else
		{
			m_VoxelRenderDataUsed = m_VoxelData;
			m_VoxelRenderTotal = Framework::s_VoxelCurrentTotal;
		}
	}

	void DODVoxels::Render()
	{
		// clear screen

		glClearColor(0.f, 0.f, 1.f, 1.f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// write visible voxels to buffers

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);
		Vec3* write_vertices = (Vec3*)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor);
		Vec3* write_color = (Vec3*)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);

		VertexData write;
		write.vertex = write_vertices;
		write.color = write_color;

		GenerateFaces(&write, m_VoxelRenderDataUsed, m_VoxelRenderTotal);

		glUnmapBuffer(GL_ARRAY_BUFFER); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);
		glUnmapBuffer(GL_ARRAY_BUFFER); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0);

		// draw buffers

		glEnableClientState(GL_VERTEX_ARRAY); CGLE();
		glEnableClientState(GL_COLOR_ARRAY); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);	glVertexPointer(3, GL_FLOAT, 0, 0); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor);		glColorPointer(3, GL_FLOAT, 0, 0); CGLE();

			glDrawArrays(GL_QUADS, 0, Framework::s_VoxelCurrentTotal * 4 * 6); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0); CGLE();

		glDisableClientState(GL_COLOR_ARRAY); CGLE();
		glDisableClientState(GL_VERTEX_ARRAY); CGLE();

		// raycast debug

		if (Framework::s_RayCast && Framework::s_Debug)
		{
			AddRayCastLines(m_VoxelRayCast, m_VoxelRayCastTotal);
		}
	}

	void DODVoxels::CloseDown()
	{

	}

	void DODVoxels::GenerateVoxels(
		VoxelData* a_OutVoxels, 
		unsigned int a_VoxelCount
	)
	{
		float last = -100;
		int side = -1;
		float sign = 1.f;

		float* dst_pos_x = a_OutVoxels->x;
		float* dst_pos_y = a_OutVoxels->y;
		float* dst_pos_z = a_OutVoxels->z;
		float* dst_color_r = a_OutVoxels->r;
		float* dst_color_g = a_OutVoxels->g;
		float* dst_color_b = a_OutVoxels->b;

		dst_pos_x[0] = dst_pos_y[0] = dst_pos_z[0] = 0.f;

		for (unsigned int i = 1; i < a_VoxelCount; i++)
		{
			// position

			dst_pos_x[i] = dst_pos_x[i - 1];
			dst_pos_y[i] = dst_pos_y[i - 1];
			dst_pos_z[i] = dst_pos_z[i - 1];

			while (side * sign == last)
			{
				side = (int)Math::Random(0.f, 3.f);
				sign = (Math::Random(0.f, 100.f) > 50.f) ? 1.f : -1.f;
			} 

			last = side * sign;

			switch (side)
			{
			case 0:
				dst_pos_x[i] += m_VoxelSize * sign;
				break;
			case 1:
				dst_pos_y[i] += m_VoxelSize * sign;
				break;
			case 2:
				dst_pos_z[i] += m_VoxelSize * sign;
				break;
			}

			// color

			dst_color_r[i] = Math::Random(0.f, 1.f);
			dst_color_g[i] = Math::Random(0.f, 1.f);
			dst_color_b[i] = Math::Random(0.f, 1.f);
		}
	}

	void DODVoxels::GenerateFaces(
		VertexData* a_OutVertices,
		VoxelData* a_InVoxels, 
		unsigned int a_Count
	)
	{
		float* src_pos_x = a_InVoxels->x;
		float* src_pos_y = a_InVoxels->y;
		float* src_pos_z = a_InVoxels->z;

		Vec3* dst_pos = a_OutVertices->vertex;

		Vec3* offset[] = {
			// Front
			&m_VoxelOffset[0], &m_VoxelOffset[1], &m_VoxelOffset[3], &m_VoxelOffset[2], 
			// Back
			&m_VoxelOffset[4], &m_VoxelOffset[6], &m_VoxelOffset[7], &m_VoxelOffset[5],
			// Top
			&m_VoxelOffset[6], &m_VoxelOffset[2], &m_VoxelOffset[3], &m_VoxelOffset[7],
			// Bottom
			&m_VoxelOffset[4], &m_VoxelOffset[5], &m_VoxelOffset[1], &m_VoxelOffset[0],
			// Left
			&m_VoxelOffset[4], &m_VoxelOffset[0], &m_VoxelOffset[2], &m_VoxelOffset[6],
			// Right
			&m_VoxelOffset[5], &m_VoxelOffset[7], &m_VoxelOffset[3], &m_VoxelOffset[1]
		};

		for (unsigned int i = 0; i < a_Count; i++)
		{
			Vec3* write_pos = dst_pos;
			Vec3** read_offset = offset;
			Vec3 pos(*src_pos_x, *src_pos_y, *src_pos_z);

			for (unsigned int j = 0; j < 24; j += 4)
			{
				write_pos[0] = pos + *read_offset[0];
				write_pos[1] = pos + *read_offset[1];
				write_pos[2] = pos + *read_offset[2];
				write_pos[3] = pos + *read_offset[3];

				write_pos += 4;
				read_offset += 4;
			}

			dst_pos += 24;

			src_pos_x++;
			src_pos_y++;
			src_pos_z++;
		}

		Vec3* dst_color = a_OutVertices->color;

		float* src_color_r = a_InVoxels->r;
		float* src_color_g = a_InVoxels->g;
		float* src_color_b = a_InVoxels->b;

		for (unsigned int i = 0; i < a_Count; i++)
		{
			Vec3 color(*src_color_r, *src_color_g, *src_color_b);

			for (unsigned int j = 0; j < 24; j += 4)
			{
				dst_color[0] = color;
				dst_color[1] = color;
				dst_color[2] = color;
				dst_color[3] = color;

				dst_color += 4;
			}

			src_color_r++;
			src_color_g++;
			src_color_b++;
		}
	}

	void DODVoxels::GenerateRayCast(
		RayCastData* a_OutRays, unsigned int& a_OutRayCount,
		VoxelData* a_InVoxels, unsigned int a_VoxelCount,
		Vec3 a_InTarget
	)
	{
		float* dst_invdir_x = a_OutRays->dx;
		float* dst_invdir_y = a_OutRays->dy;
		float* dst_invdir_z = a_OutRays->dz;
		float* dst_pos_x = a_OutRays->ox;
		float* dst_pos_y = a_OutRays->oy;
		float* dst_pos_z = a_OutRays->oz;

		float* src_pos_x = a_InVoxels->x;
		float* src_pos_y = a_InVoxels->y;
		float* src_pos_z = a_InVoxels->z;

		unsigned int total = 0;

		for (unsigned int i = 0; i < a_VoxelCount; i++)
		{
			float x = *src_pos_x++ - a_InTarget.x;
			float y = *src_pos_y++ - a_InTarget.y;
			float z = *src_pos_z++ - a_InTarget.z;

			*dst_pos_x++ = a_InTarget.x;
			*dst_pos_y++ = a_InTarget.y;
			*dst_pos_z++ = a_InTarget.z;

			float rcp_length = 1.f / Math::Sqrt((x * x) + (y * y) + (z * z));

			*dst_invdir_x++ = 1.f / (x * rcp_length);
			*dst_invdir_y++ = 1.f / (y * rcp_length);
			*dst_invdir_z++ = 1.f / (z * rcp_length);

			total++;
		}

		a_OutRayCount = total;
	}

	void DODVoxels::SolveRayCast(
		VoxelData* a_OutVoxels, unsigned int& a_OutVoxelCount,
		RayCastData* a_InRays, unsigned int a_InRayCount, 
		VoxelData* a_InVoxels, unsigned int a_InVoxelCount
	)
	{
		float* dst_voxel_pos_x = a_OutVoxels->x;
		float* dst_voxel_pos_y = a_OutVoxels->y;
		float* dst_voxel_pos_z = a_OutVoxels->z;
		float* dst_voxel_color_r = a_OutVoxels->r;
		float* dst_voxel_color_g = a_OutVoxels->g;
		float* dst_voxel_color_b = a_OutVoxels->b;
		
		float* src_ray_pos_x = a_InRays->ox;
		float* src_ray_pos_y = a_InRays->oy;
		float* src_ray_pos_z = a_InRays->oz;
		float* src_ray_dir_x = a_InRays->dx;
		float* src_ray_dir_y = a_InRays->dy;
		float* src_ray_dir_z = a_InRays->dz;

		float minsize = -(m_VoxelSize / 2.f);
		float maxsize =  (m_VoxelSize / 2.f);

		unsigned int total = 0;

		for (unsigned int i = 0; i < a_InRayCount; i++)
		{
			float* src_voxel_pos_x = a_InVoxels->x;
			float* src_voxel_pos_y = a_InVoxels->y;
			float* src_voxel_pos_z = a_InVoxels->z;

			float min_time = 10000.f;
			unsigned int index = 0;

			float ray_precomp_pos_x = (*src_ray_pos_x) * (*src_ray_dir_x);
			float ray_precomp_pos_y = (*src_ray_pos_y) * (*src_ray_dir_y);
			float ray_precomp_pos_z = (*src_ray_pos_z) * (*src_ray_dir_z);

			float ray_dir_x_min = minsize * (*src_ray_dir_x);
			float ray_dir_y_min = minsize * (*src_ray_dir_y);
			float ray_dir_z_min = minsize * (*src_ray_dir_z);
			float ray_dir_x_max = maxsize * (*src_ray_dir_x);
			float ray_dir_y_max = maxsize * (*src_ray_dir_y);
			float ray_dir_z_max = maxsize * (*src_ray_dir_z);

			bool hit = false;

			for (unsigned int j = 0; j < a_InVoxelCount; j++)
			{
				float tmin = 0.f;
				float tmax = 10000.f;

				float t0, t1;

				float precompute_pos_x = ((*src_voxel_pos_x) * (*src_ray_dir_x)) - ray_precomp_pos_x;

				t0 = ray_dir_x_min + precompute_pos_x;
				t1 = ray_dir_x_max + precompute_pos_x;
				if (t0 > t1)
				{
					tmin = (t1 > tmin) ? t1 : tmin;
					tmax = (t0 < tmax) ? t0 : tmax;
				}
				else
				{
					tmin = (t0 > tmin) ? t0 : tmin;
					tmax = (t1 < tmax) ? t1 : tmax;
				}

				float precompute_pos_y = ((*src_voxel_pos_y) * (*src_ray_dir_y)) - ray_precomp_pos_y;

				t0 = ray_dir_y_min + precompute_pos_y;
				t1 = ray_dir_y_max + precompute_pos_y;
				if (t0 > t1)
				{
					tmin = (t1 > tmin) ? t1 : tmin;
					tmax = (t0 < tmax) ? t0 : tmax;
				}
				else
				{
					tmin = (t0 > tmin) ? t0 : tmin;
					tmax = (t1 < tmax) ? t1 : tmax;
				}

				float precompute_pos_z = ((*src_voxel_pos_z) * (*src_ray_dir_z)) - ray_precomp_pos_z;

				t0 = ray_dir_z_min + precompute_pos_z;
				t1 = ray_dir_z_max + precompute_pos_z;
				if (t0 > t1)
				{
					tmin = (t1 > tmin) ? t1 : tmin;
					tmax = (t0 < tmax) ? t0 : tmax;
				}
				else
				{
					tmin = (t0 > tmin) ? t0 : tmin;
					tmax = (t1 < tmax) ? t1 : tmax;
				}

				if (tmin <= tmax && tmin <= min_time)
				{
					index = j;

					min_time = tmin;

					hit = true;
				}

				src_voxel_pos_x++;
				src_voxel_pos_y++;
				src_voxel_pos_z++;
			}
			
			if (hit) 
			{ 
				*dst_voxel_pos_x++ = a_InVoxels->x[index];
				*dst_voxel_pos_y++ = a_InVoxels->y[index];
				*dst_voxel_pos_z++ = a_InVoxels->z[index];

				*dst_voxel_color_r++ = a_InVoxels->r[index];
				*dst_voxel_color_g++ = a_InVoxels->g[index];
				*dst_voxel_color_b++ = a_InVoxels->b[index];

				total++; 
			}

			src_ray_pos_x++;
			src_ray_pos_y++;
			src_ray_pos_z++;

			src_ray_dir_x++;
			src_ray_dir_y++;
			src_ray_dir_z++;
		}

		a_OutVoxelCount = total;
	}

	unsigned int DODVoxels::AddRayCastLines(
		RayCastData* a_InRays, unsigned int a_RayCount
	)
	{
		float* src_pos_x = a_InRays->ox;
		float* src_pos_y = a_InRays->oy;
		float* src_pos_z = a_InRays->oz;

		float* src_dir_x = a_InRays->dx;
		float* src_dir_y = a_InRays->dy;
		float* src_dir_z = a_InRays->dz;

		glColor3f(1.f, 1.f, 0.f);

		glBegin(GL_LINES);

		for (unsigned int i = 0; i < a_RayCount; i++)
		{
			glVertex3f(
				*src_pos_x, 
				*src_pos_y, 
				*src_pos_z
			);
			glVertex3f(
				*src_pos_x + (1.f / *src_dir_x++) * 100.f, 
				*src_pos_y + (1.f / *src_dir_y++) * 100.f, 
				*src_pos_z + (1.f / *src_dir_z++) * 100.f
			);

			src_pos_x++;
			src_pos_y++;
			src_pos_z++;
		}
		glEnd();

		return 0;
	}

}