#ifndef _DODVOXELS_H_
#define _DODVOXELS_H_

#include "Game.h"

namespace Exp
{

	class DODVoxels : public Game
	{

		struct VoxelData
		{
			float* x;
			float* y;
			float* z;

			float* r;
			float* g;
			float* b;
		};

		struct RayCastData
		{
			float* ox;
			float* oy;
			float* oz;

			float* dx;
			float* dy;
			float* dz;
		};

		struct VertexData
		{
			Vec3* vertex;
			Vec3* color;
		};
	
	public:
	
		void Init();
		void Tick(float a_DT);
		void Reload();
		void PostTick();
		void Render();
		void CloseDown();

		void GenerateRayCast(
			RayCastData* a_OutRays, unsigned int& a_OutRayCount,
			VoxelData* a_InVoxels, unsigned int a_VoxelCount,
			Vec3 a_InTarget
		);

		void SolveRayCast(
			VoxelData* a_OutVoxels, unsigned int& a_OutVoxelCount,
			RayCastData* a_InRays, unsigned int a_InRayCount,
			VoxelData* a_InVoxels, unsigned int a_InVoxelCount
		);

		unsigned int AddRayCastLines(
			RayCastData* a_InRays, unsigned int a_RayCount
		);

		void GenerateVoxels(
			VoxelData* a_OutVoxels, unsigned int a_VoxelCount
		);

		void GenerateFaces(
			VertexData* a_OutVertices,
			VoxelData* a_InVoxels, unsigned int a_Count
		);

	private:

		float m_VoxelSize;
		Vec3 m_VoxelOffset[8];

		unsigned int m_VoxelCurrentTotal;
		unsigned int m_VoxelTotal;

		VoxelData* m_VoxelData;
		unsigned int m_VoxelDataSize;

		RayCastData* m_VoxelRayCast;
		unsigned int m_VoxelRayCastTotal;

		VoxelData* m_VoxelRayCastData;
		unsigned int m_VoxelRayCastDataTotal;

		VoxelData* m_VoxelRenderData;
		VoxelData* m_VoxelRenderDataUsed;
		unsigned int m_VoxelRenderTotal;

		Vec3* m_VoxelVertices;
		Vec3* m_VoxelColors;
		
		Vec3* m_VoxelDebugRayCast;
		unsigned int m_VoxelDebugRayCastTotal;

		GLuint m_VoxelVBOVertex;
		GLuint m_VoxelVBOColor;
	
	}; // class DODVoxels

}; // namespace Exp
	
#endif