#include "Common.h"

#include "Framework.h"

#include "Game.h"
#include "DODVoxels.h"
#include "OOPVoxels.h"

namespace Exp
{

	Game* g_GameDOD = NULL;
	Game* g_GameOOP = NULL;
	Game* g_GameCurrent = NULL;

	unsigned int Framework::s_VoxelTotal = 100000;
	unsigned int Framework::s_VoxelCurrentTotal = 2000;
	bool Framework::s_Debug = true;
	bool Framework::s_RayCast = true;
	bool Framework::s_TargetCamera = false;
	float Framework::s_CameraRotation = 0.f;
	float Framework::s_TargetRotation = 0.f;
	bool Framework::s_KeysPressed[256];
	bool Framework::s_KeysReleased[256];

	static unsigned long g_TimeCurrent, g_TimeStart;
	static float         g_TimeDelta;
	static float         g_PhysicsTime = 0;
	static float         g_RenderTime = 0;

	bool         g_Exit = false;
	bool         g_Active = true;
	HINSTANCE    g_Instance;
	HWND         g_Window;
	MSG          g_MSG;
	HDC          g_WindowContext;

	static GLuint       g_FontBase = 1000;

	static LARGE_INTEGER g_TimeFPSStart, g_TimeFPSEnd;
	static float g_TimeFPSFrequency;
	float g_StackFPS = 0.f;
	float g_AverageFPS = 0.f;
	float g_LowestFPS = 1000.f;
	float g_HighestFPS = -1000.f;
	int g_CountFPS = 0;
	bool g_ChangeRender = false;

	const float frame_physics = (1000.f / 60.f);

	LRESULT CALLBACK Win32Messages(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
	{
		switch (message)
		{

		case WM_ACTIVATE:
			{
				g_Active = ((LOWORD(wParam) != WA_INACTIVE) && !((BOOL)HIWORD(wParam)));
				break;
			}

		case WM_KEYDOWN:
			{
				Framework::s_KeysPressed[wParam] = true;

				if (wParam == VK_ESCAPE) { g_Exit = true; }

				break;
			}

		case WM_KEYUP:
			{
				Framework::s_KeysPressed[wParam] = false;
				Framework::s_KeysReleased[wParam] = true;

				break;
			}

		case WM_MOUSEMOVE:
			{

				break;
			}

		case WM_CLOSE:
		case WM_QUIT:
		case WM_DESTROY:
			g_Exit = true;
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);

		}

		return 0;
	}

	void InitCamera()
	{		
		glMatrixMode(GL_PROJECTION);
		gluPerspective(60.f, (float)EXP_SCRWIDTH / (float)EXP_SCRHEIGHT, 0.1f, 10000.f);

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
	}

	void TickCamera(float a_DT)
	{
		if (Framework::s_KeysPressed[VK_LEFT]) { Framework::s_CameraRotation += 3.0f * a_DT; }
		if (Framework::s_KeysPressed[VK_RIGHT]) { Framework::s_CameraRotation -= 3.0f * a_DT; }

		Framework::s_TargetRotation += 1.f * a_DT;
	}

	void RenderCamera()
	{
		Vec3 campos;

		if (Framework::s_TargetCamera)
		{
			campos.x = Math::CosDeg(Framework::s_TargetRotation) * 500.f;
			campos.y = 10.f;
			campos.z = Math::SinDeg(Framework::s_TargetRotation) * 500.f;
		}
		else
		{
			campos.x = Math::CosDeg(Framework::s_CameraRotation) * 1000.f;
			campos.y = 100.f;
			campos.z = Math::SinDeg(Framework::s_CameraRotation) * 1000.f;
		}

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		gluLookAt(
			campos.x, campos.y, campos.z,
			0.f, 0.f, 0.f,
			0.f, 1.f, 0.f
		);
	}

}; // namespace Exp

using namespace Exp;

int main(int argc, char** argv)
{
	LOG("Starting program.");

	WNDCLASS temp = { 
		CS_HREDRAW | CS_VREDRAW | CS_OWNDC, // defines the class style 
		Win32Messages,	                    // pointer to the message handler function
		0,                                  // extra bytes following the window class structure
		0,                                  // extra bytes following the window instance
		g_Instance,                         // instance handle				
		LoadIcon(NULL, IDI_APPLICATION),    // icon handle
		LoadCursor(NULL, IDC_ARROW),        // cursor
		NULL,                               // background brush
		NULL,                               // menu name
		EXP_WINDOWTITLE                     // window title
	};
	if (!RegisterClass(&temp)) { return 0; }

	// center windows on screen

	const int x = GetSystemMetrics(SM_CXSCREEN) / 2 - (EXP_SCRWIDTH / 2);
	const int y = GetSystemMetrics(SM_CYSCREEN) / 2 - (EXP_SCRHEIGHT / 2);

	const DWORD style = WS_OVERLAPPEDWINDOW & ~(WS_THICKFRAME | WS_MAXIMIZEBOX);
	RECT dim = { 
		0, 0, 
		EXP_SCRWIDTH, EXP_SCRHEIGHT 
	};
	AdjustWindowRect(&dim, style, FALSE);
	const int w = dim.right - dim.left;
	const int h = dim.bottom - dim.top;

	g_Window = CreateWindowEx (
		0,
		EXP_WINDOWTITLE,       // window class
		EXP_WINDOWTITLE,   // window title
		style,             // visibility settings
		x, y,
		w, h,
		NULL, 
		NULL, 
		GetModuleHandle(NULL),
		NULL
	);
	if (!g_Window)
	{ 
		return -1; 
	}

	LOG("Created window.");

	if (!(g_WindowContext = GetDC(g_Window))) 
	{ 
		return -1; 
	}

	LOG("Retrieved window context.");

	PIXELFORMATDESCRIPTOR pfd = { 
		sizeof(PIXELFORMATDESCRIPTOR), 
		1, 
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA, 
		32, 
		0, 0, 0, 0, 0, 0, 
		8, 
		0, 
		0, 0, 0, 0, 
		0, 
		24, 
		0, 
		0,
		PFD_MAIN_PLANE,	
		0, 
		0, 0, 0
	};

	unsigned int format;
	if (!(format = ChoosePixelFormat(g_WindowContext, &pfd))) 
	{ 
		return -1; 
	}

	LOG("Chose pixel format.");

	if (!SetPixelFormat(g_WindowContext, format, &pfd)) 
	{ 
		return -1; 
	}

	LOG("Set pixel format.");

	HGLRC hrc;
	if (!(hrc = wglCreateContext(g_WindowContext)))
	{ 
		return -1; 
	}

	LOG("Created OpenGL context.");

	if (!wglMakeCurrent(g_WindowContext, hrc))
	{ 
		return -1; 
	}

	LOG("Made context current.");

	// initialize opengl

	GLenum status = glewInit();
	if (status != GLEW_OK)
	{
		char error[256];
		sprintf(error, "%s", glewGetErrorString(status));
		MessageBoxA(NULL, error, "GLEW error", MB_OK | MB_ICONEXCLAMATION);
		return -1;
	}

	LOG("Initialized GLEW.");

	// turn off vsync

	wglSwapIntervalEXT(0);

	LOG("Turned off vsync.");

	// clear keyboard

	for (int i = 0; i < 256; i++) 
	{ 
		Framework::s_KeysPressed[i] = false; 
		Framework::s_KeysReleased[i] = false; 
	}

	LOG("Initialized keyboard.");

	// initialize implementations

	unsigned int seed = (unsigned int)(time(0));

	Math::Seed(seed);

	g_GameDOD = new DODVoxels;
	g_GameDOD->Init();

	LOG("Initialized DOD version.");

	Math::Seed(seed);

	g_GameOOP = new OOPVoxels;
	g_GameOOP->Init();

	LOG("Initialized OOP version.");

	g_GameCurrent = g_GameDOD;

	// show window

	ShowWindow(g_Window, 1);
	UpdateWindow(g_Window);

	LOG("Showed window.");

	// initialize font

	HFONT oldfont;

	g_FontBase = glGenLists(96);					
	HFONT font_text = CreateFont(
		24,
		0,
		0,						// Angle Of Escapement
		0,
		FW_BOLD,				// bold
		FALSE,					// italic
		FALSE,					// underline
		FALSE,					// strikeout
		ANSI_CHARSET,	
		OUT_TT_PRECIS,	
		CLIP_DEFAULT_PRECIS,
		ANTIALIASED_QUALITY,
		FF_DONTCARE | DEFAULT_PITCH,	
		"Courier New"
	);	
	oldfont = (HFONT)SelectObject(g_WindowContext, font_text);
	wglUseFontBitmaps(g_WindowContext, 0, 127, g_FontBase);

	LOG("Created font display list.");

	InitCamera();

	// start loop

	QueryPerformanceFrequency(&g_TimeFPSEnd);
	g_TimeFPSFrequency = 1000.f / (float)g_TimeFPSEnd.QuadPart;
	QueryPerformanceCounter(&g_TimeFPSStart);

	g_TimeStart = GetTickCount();

	LOG("Starting loop.");

	while (1)
	{
		if (PeekMessage(&g_MSG, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&g_MSG);
			DispatchMessage(&g_MSG);
		}
		else
		{
			g_TimeCurrent = GetTickCount();
			g_TimeDelta = (float)(g_TimeCurrent - g_TimeStart);
			g_TimeStart = g_TimeCurrent;

			g_PhysicsTime += g_TimeDelta;

			g_ChangeRender = false;

			while (g_PhysicsTime > frame_physics)
			{
				TickCamera(1.f);

				// framework update

				if (Framework::s_KeysReleased['D'])
				{
					Framework::s_Debug = !Framework::s_Debug;
				}

				if (Framework::s_KeysReleased['R'])
				{
					Framework::s_RayCast = !Framework::s_RayCast;
					g_ChangeRender = true;
				}

				if (Framework::s_KeysReleased['Q'])
				{
					Framework::s_VoxelCurrentTotal -= 1000;
					if ((int)Framework::s_VoxelCurrentTotal <= 0) { Framework::s_VoxelCurrentTotal = 100; }

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				if (Framework::s_KeysReleased['W'])
				{
					Framework::s_VoxelCurrentTotal += 1000;
					if (Framework::s_VoxelCurrentTotal >= Framework::s_VoxelTotal) { Framework::s_VoxelCurrentTotal = Framework::s_VoxelTotal; }

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				if (Framework::s_KeysReleased['1'])
				{
					Framework::s_VoxelCurrentTotal = 4000;

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				if (Framework::s_KeysReleased['2'])
				{
					Framework::s_VoxelCurrentTotal = 90000;

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				if (Framework::s_KeysReleased['K'])
				{
					Framework::s_TargetCamera = !Framework::s_TargetCamera;
				}

				if (Framework::s_KeysReleased['G'])
				{
					unsigned int seed = (unsigned int)(time(0));

					Math::Seed(seed);
					g_GameDOD->Reload();

					Math::Seed(seed);
					g_GameOOP->Reload();

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				if (Framework::s_KeysReleased[VK_SPACE])
				{
					if (g_GameCurrent == g_GameDOD) { g_GameCurrent = g_GameOOP; }
					else if (g_GameCurrent == g_GameOOP) { g_GameCurrent = g_GameDOD; }

					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;
				}

				g_GameCurrent->Tick(1.f);

				for (int i = 0; i < 256; i++) { Framework::s_KeysReleased[i] = false; }

				g_PhysicsTime -= frame_physics;
			}

			g_GameCurrent->PostTick();

			if (g_Active)
			{
				RenderCamera();

				g_GameCurrent->Render();

				if (g_ChangeRender)
				{
					g_LowestFPS = 1000.f;
					g_HighestFPS = -1000.f;

					QueryPerformanceCounter(&g_TimeFPSStart);
					g_CountFPS = 0;

					g_ChangeRender = false;
				}

				// render fps
				if (++g_CountFPS == 10)
				{
					//unsigned int curr = timeGetTime();
					QueryPerformanceCounter(&g_TimeFPSEnd);
					float samples = (float)(g_TimeFPSEnd.QuadPart - g_TimeFPSStart.QuadPart) / 10.f * g_TimeFPSFrequency;

					g_AverageFPS = 1000.f / samples;
					g_LowestFPS = (g_AverageFPS < g_LowestFPS) ? g_AverageFPS : g_LowestFPS;
					g_HighestFPS = (g_AverageFPS > g_HighestFPS) ? g_AverageFPS : g_HighestFPS;

					g_TimeFPSStart = g_TimeFPSEnd;

					g_CountFPS = 0;
				}

				char msg[256];

				glViewport(0, 0, EXP_SCRWIDTH, EXP_SCRHEIGHT);

				glMatrixMode(GL_PROJECTION);
				glPushMatrix();
					glLoadIdentity();
					gluOrtho2D(
						0, EXP_SCRWIDTH, 
						EXP_SCRHEIGHT, 0
					);

					glMatrixMode(GL_MODELVIEW);
					glPushMatrix();
						glLoadIdentity();
							
							glPushAttrib(GL_LIST_BIT);   
								glListBase(g_FontBase); 

								glRasterPos3i(10, 30, 0);
								sprintf(msg, "FPS Avg: %.f (Lo: %.f Hi: %.f)", g_AverageFPS, g_LowestFPS, g_HighestFPS);
								glCallLists(strlen(msg), GL_UNSIGNED_BYTE, msg);  

								glRasterPos3i(10, 60, 0);
								sprintf(msg, "Voxels: %i", Framework::s_VoxelCurrentTotal);
								glCallLists(strlen(msg), GL_UNSIGNED_BYTE, msg);  

								glRasterPos3i(10, 90, 0);
								sprintf(msg, "Method: %s", (g_GameCurrent == g_GameDOD) ? "DOD" : "OOP");
								glCallLists(strlen(msg), GL_UNSIGNED_BYTE, msg);  

							glPopAttrib();
					glPopMatrix();

				glMatrixMode(GL_PROJECTION);
				glPopMatrix();
				glMatrixMode(GL_MODELVIEW);

				SwapBuffers(g_WindowContext);
			}
		}

		if (g_Exit) { break; }
	}

	g_GameDOD->CloseDown();
	g_GameOOP->CloseDown();

	DestroyWindow(g_Window);

	return 0;
}