#include "Common.h"

#include "Framework.h"
#include "OOPVoxels.h"

namespace Exp
{

	OOPVoxels::Voxel::Voxel()
	{
		m_Clipped = false;
	}

	OOPVoxels::Voxel::~Voxel()
	{

	}

	void OOPVoxels::Voxel::SetPosition(Vec3& a_Position)
	{
		m_Pos = a_Position;

		Vec3 size(m_Size / 2.f, m_Size / 2.f, m_Size / 2.f);
		m_Min = m_Pos - size;
		m_Max = m_Pos + size;
	}

	void OOPVoxels::Voxel::SetSize(float a_Size)
	{
		m_Size = a_Size;
	}

	bool OOPVoxels::Voxel::IsClipped()
	{
		return m_Clipped;
	}

	void OOPVoxels::Voxel::Start()
	{
		m_Clipped = false;
	}

	void OOPVoxels::Voxel::GenerateRay(Vec3 a_Target)
	{
		Vec3 dir =  (m_Pos - a_Target).Normalize();
		m_RayDirection = dir.Reciprocral();
		m_RayOrigin = a_Target;
		m_MinTime = 0.f;
		m_MaxTime = 10000.f;
	}

	bool OOPVoxels::Voxel::RayCollides(Voxel* a_Other)
	{
		float t0, t1;

		t0 = (a_Other->GetMin().x - m_RayOrigin.x) * m_RayDirection.x;
		t1 = (a_Other->GetMax().x - m_RayOrigin.x) * m_RayDirection.x;
		if (t0 > t1)
		{
			m_MinTime = (t1 > m_MinTime) ? t1 : m_MinTime;
			m_MaxTime = (t0 < m_MaxTime) ? t0 : m_MaxTime;
		}
		else
		{
			m_MinTime = (t0 > m_MinTime) ? t0 : m_MinTime;
			m_MaxTime = (t1 < m_MaxTime) ? t1 : m_MaxTime;
		}

		t0 = (a_Other->GetMin().y - m_RayOrigin.y) * m_RayDirection.y;
		t1 = (a_Other->GetMax().y - m_RayOrigin.y) * m_RayDirection.y;
		if (t0 > t1)
		{
			m_MinTime = (t1 > m_MinTime) ? t1 : m_MinTime;
			m_MaxTime = (t0 < m_MaxTime) ? t0 : m_MaxTime;
		}
		else
		{
			m_MinTime = (t0 > m_MinTime) ? t0 : m_MinTime;
			m_MaxTime = (t1 < m_MaxTime) ? t1 : m_MaxTime;
		}

		t0 = (a_Other->GetMin().z - m_RayOrigin.z) * m_RayDirection.z;
		t1 = (a_Other->GetMax().z - m_RayOrigin.z) * m_RayDirection.z;
		if (t0 > t1)
		{
			m_MinTime = (t1 > m_MinTime) ? t1 : m_MinTime;
			m_MaxTime = (t0 < m_MaxTime) ? t0 : m_MaxTime;
		}
		else
		{
			m_MinTime = (t0 > m_MinTime) ? t0 : m_MinTime;
			m_MaxTime = (t1 < m_MaxTime) ? t1 : m_MaxTime;
		}

		if (m_MinTime <= m_MaxTime)
		{
			m_Clipped = false;
			m_Hit = a_Other;
		}

		return false;
	}

	void OOPVoxels::Voxel::AddToRenderList(Vec3* a_Vertex, Vec3* a_Color, Vec3* a_Offset)
	{
		Vec3* dst_pos = a_Vertex;
		Vec3* dst_color = a_Color;

		Vec3* offset[] = {
			// Front
			&a_Offset[0], &a_Offset[1], &a_Offset[3], &a_Offset[2], 
			// Back
			&a_Offset[4], &a_Offset[6], &a_Offset[7], &a_Offset[5],
			// Top
			&a_Offset[6], &a_Offset[2], &a_Offset[3], &a_Offset[7],
			// Bottom
			&a_Offset[4], &a_Offset[5], &a_Offset[1], &a_Offset[0],
			// Left
			&a_Offset[4], &a_Offset[0], &a_Offset[2], &a_Offset[6],
			// Right
			&a_Offset[5], &a_Offset[7], &a_Offset[3], &a_Offset[1]
		};

		Vec3** read_offset = offset;

		for (unsigned int j = 0; j < 24; j += 4)
		{
			dst_pos[0] = m_Pos + *read_offset[0];
			dst_pos[1] = m_Pos + *read_offset[1];
			dst_pos[2] = m_Pos + *read_offset[2];
			dst_pos[3] = m_Pos + *read_offset[3];

			dst_color[0] = m_Color;
			dst_color[1] = m_Color;
			dst_color[2] = m_Color;
			dst_color[3] = m_Color;

			dst_color += 4;
			dst_pos += 4;
			read_offset += 4;
		}
	}

	OOPVoxels::Ray::Ray()
	{

	}

	OOPVoxels::Ray::~Ray()
	{

	}

	void OOPVoxels::Ray::SetOrigin(Vec3& a_Origin)
	{
		m_Origin = a_Origin;
	}

	void OOPVoxels::Ray::SetDirection(Vec3& a_Direction)
	{
		m_Direction = a_Direction;
	}

	bool OOPVoxels::Ray::CollidesWith(Voxel* a_Other)
	{
		float t0, t1;
		float tmin = 0.f;
		float tmax = 10000.f;

		t0 = (a_Other->GetMin().x - m_Origin.x) * m_Direction.x;
		t1 = (a_Other->GetMax().x - m_Origin.x) * m_Direction.x;
		if (t0 > t1)
		{
			tmin = (t1 > tmin) ? t1 : tmin;
			tmax = (t0 < tmax) ? t0 : tmax;
		}
		else
		{
			tmin = (t0 > tmin) ? t0 : tmin;
			tmax = (t1 < tmax) ? t1 : tmax;
		}

		t0 = (a_Other->GetMin().y - m_Origin.y) * m_Direction.y;
		t1 = (a_Other->GetMax().y - m_Origin.y) * m_Direction.y;
		if (t0 > t1)
		{
			tmin = (t1 > tmin) ? t1 : tmin;
			tmax = (t0 < tmax) ? t0 : tmax;
		}
		else
		{
			tmin = (t0 > tmin) ? t0 : tmin;
			tmax = (t1 < tmax) ? t1 : tmax;
		}

		t0 = (a_Other->GetMin().z - m_Origin.z) * m_Direction.z;
		t1 = (a_Other->GetMax().z - m_Origin.z) * m_Direction.z;
		if (t0 > t1)
		{
			tmin = (t1 > tmin) ? t1 : tmin;
			tmax = (t0 < tmax) ? t0 : tmax;
		}
		else
		{
			tmin = (t0 > tmin) ? t0 : tmin;
			tmax = (t1 < tmax) ? t1 : tmax;
		}

		if (tmin <= tmax && tmin <= m_TimeMin)
		{
			m_TimeMin = tmin;
			m_Hit = a_Other;

			return true;
		}

		return false;
	}

	OOPVoxels::Voxel* OOPVoxels::Ray::GetHit()
	{
		//if (m_Owner == m_Hit) { return m_Hit; }
		//return NULL;
		return m_Hit;
	}

	void OOPVoxels::Ray::Begin()
	{
		m_TimeMin = 100000.f;
		m_TimeMax = 0.f;
		m_Hit = NULL;
		m_Owner = NULL;
	}

	void OOPVoxels::Ray::SetOwner(Voxel* a_Owner)
	{
		m_Owner = a_Owner;
	}

	void OOPVoxels::Init()
	{
		m_VoxelSize = 20.f;

		m_VoxelDataSize = Framework::s_VoxelTotal * 4 * 6;

		m_Voxels = new Voxel[Framework::s_VoxelTotal];
		for (unsigned int i = 0; i < Framework::s_VoxelTotal; i++)
		{
			m_Voxels[i].SetSize(m_VoxelSize);
		}

		m_Rays = new Ray[Framework::s_VoxelTotal];

		m_VoxelVertices = new Vec3[m_VoxelDataSize];
		m_VoxelColors = new Vec3[m_VoxelDataSize];

		// generate voxels

		GenerateVoxels();

		// offsets

		float halfsize = m_VoxelSize / 2.f;

		m_VoxelOffset[0] = Vec3(-halfsize, -halfsize,  halfsize);
		m_VoxelOffset[1] = Vec3( halfsize, -halfsize,  halfsize);
		m_VoxelOffset[2] = Vec3(-halfsize,  halfsize,  halfsize);
		m_VoxelOffset[3] = Vec3( halfsize,  halfsize,  halfsize);

		m_VoxelOffset[4] = Vec3(-halfsize, -halfsize, -halfsize);
		m_VoxelOffset[5] = Vec3( halfsize, -halfsize, -halfsize);
		m_VoxelOffset[6] = Vec3(-halfsize,  halfsize, -halfsize);
		m_VoxelOffset[7] = Vec3( halfsize,  halfsize, -halfsize);

		// voxel buffer

		glGenBuffers(1, &m_VoxelVBOVertex); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex); CGLE();
		glBufferData(GL_ARRAY_BUFFER, m_VoxelDataSize * sizeof(Vec3), m_VoxelVertices, GL_STREAM_DRAW); CGLE();

		glGenBuffers(1, &m_VoxelVBOColor); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor); CGLE();
		glBufferData(GL_ARRAY_BUFFER, m_VoxelDataSize * sizeof(Vec3), m_VoxelColors, GL_STREAM_DRAW); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0); CGLE();
	}

	void OOPVoxels::Tick(float a_DT)
	{

	}

	void OOPVoxels::Reload()
	{
		GenerateVoxels();
	}

	void OOPVoxels::PostTick()
	{
		if (Framework::s_RayCast)
		{
			Vec3 target(
				Math::CosDeg(Framework::s_TargetRotation) * 500.f, 
				10.f, 
				Math::SinDeg(Framework::s_TargetRotation) * 500.f
			);

			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				m_Rays[i].Begin();

				m_Rays[i].SetOwner(&m_Voxels[i]);
				m_Rays[i].SetOrigin(target);
				Vec3 dir = (m_Voxels[i].GetPosition() - target).Normalize();
				m_Rays[i].SetDirection(dir.Reciprocral());
			}

			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				for (unsigned int j = 0; j < Framework::s_VoxelCurrentTotal; j++)
				{
					m_Rays[i].CollidesWith(&m_Voxels[j]);
				}
			}
		}
		else
		{
			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				m_Voxels[i].SetClipped(false);
			}
		}
	}

	void OOPVoxels::Render()
	{
		// clear screen

		glClearColor(1.f, 0.f, 0.f, 1.f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// write visible voxels to buffers

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);
		Vec3* write_vertices = (Vec3*)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor);
		Vec3* write_color = (Vec3*)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);

		unsigned int total = 0;

		if (Framework::s_RayCast)
		{
			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				if (m_Rays[i].GetHit())
				{
					m_Rays[i].GetHit()->AddToRenderList(write_vertices, write_color, m_VoxelOffset);

					write_vertices += 24;
					write_color += 24;

					total++;
				}
			}
		}
		else
		{
			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				if (!m_Voxels[i].IsClipped())
				{
					m_Voxels[i].AddToRenderList(write_vertices, write_color, m_VoxelOffset);

					write_vertices += 24;
					write_color += 24;

					total++;
				}
			}
		}
		

		glUnmapBuffer(GL_ARRAY_BUFFER); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);
		glUnmapBuffer(GL_ARRAY_BUFFER); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0);

		// draw buffers

		glEnableClientState(GL_VERTEX_ARRAY); CGLE();
		glEnableClientState(GL_COLOR_ARRAY); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOVertex);	glVertexPointer(3, GL_FLOAT, 0, 0); CGLE();
		glBindBuffer(GL_ARRAY_BUFFER, m_VoxelVBOColor);		glColorPointer(3, GL_FLOAT, 0, 0); CGLE();

			glDrawArrays(GL_QUADS, 0, total * 4 * 6); CGLE();

		glBindBuffer(GL_ARRAY_BUFFER, 0); CGLE();

		glDisableClientState(GL_COLOR_ARRAY); CGLE();
		glDisableClientState(GL_VERTEX_ARRAY); CGLE();

		if (Framework::s_RayCast && Framework::s_Debug)
		{
			glColor3f(1.f, 1.f, 0.f);

			glBegin(GL_LINES);

			for (unsigned int i = 0; i < Framework::s_VoxelCurrentTotal; i++)
			{
				/*glVertex3f(
					m_Rays[i].GetOrigin().x, 
					m_Rays[i].GetOrigin().y, 
					m_Rays[i].GetOrigin().z
				);
				glVertex3f(
					m_Rays[i].GetOrigin().x + (1.f / m_Rays[i].GetDirection().x) * 100.f, 
					m_Rays[i].GetOrigin().y + (1.f / m_Rays[i].GetDirection().y) * 100.f, 
					m_Rays[i].GetOrigin().z + (1.f / m_Rays[i].GetDirection().z) * 100.f
				);*/

				if (m_Rays[i].GetHit())
				{
					glVertex3f(
						m_Rays[i].GetHit()->GetPosition().x, 
						m_Rays[i].GetHit()->GetPosition().y, 
						m_Rays[i].GetHit()->GetPosition().z
					);
					glVertex3f(
						m_Rays[i].GetOrigin().x, 
						m_Rays[i].GetOrigin().y, 
						m_Rays[i].GetOrigin().z
					);
				}
			}

			glEnd();
		}
	}

	void OOPVoxels::CloseDown()
	{

	}

	void OOPVoxels::GenerateVoxels()
	{
		float last = -100;
		int side = -1;
		float sign = 1.f;

		m_Voxels[0].SetPosition(Vec3(0, 0, 0));

		for (unsigned int i = 1; i < Framework::s_VoxelTotal; i++)
		{
			// position

			m_Voxels[i].SetPosition(m_Voxels[i - 1].GetPosition());

			while (side * sign == last)
			{
				side = (int)Math::Random(0.f, 3.f);
				sign = (Math::Random(0.f, 100.f) > 50.f) ? 1.f : -1.f;
			} 

			last = side * sign;

			switch (side)
			{
			case 0:
				m_Voxels[i].GetPosition().x += m_VoxelSize * sign;
				break;
			case 1:
				m_Voxels[i].GetPosition().y += m_VoxelSize * sign;
				break;
			case 2:
				m_Voxels[i].GetPosition().z += m_VoxelSize * sign;
				break;
			}

			// color

			Vec3 color(
				Math::Random(0.f, 1.f),
				Math::Random(0.f, 1.f),
				Math::Random(0.f, 1.f)
			);

			m_Voxels[i].SetColor(color);
		}
	}

}