#ifndef _OOPVOXELS_H_
#define _OOPVOXELS_H_

#include "Game.h"

namespace Exp
{

	class OOPVoxels : public Game
	{

		class Voxel
		{

		public:

			Voxel();
			~Voxel();

			void SetSize(float a_Size);

			void Start();

			void GenerateRay(Vec3 a_Target);
			bool RayCollides(Voxel* a_Other);

			void SetClipped(bool a_State) { m_Clipped = a_State; }
			bool IsClipped();

			void AddToRenderList(Vec3* a_Vertex, Vec3* a_Color, Vec3* a_Offsets);

			void SetColor(Vec3& a_Color) { m_Color = a_Color; }

			Vec3& GetPosition() { return m_Pos; }
			void SetPosition(Vec3& a_Position);

			Vec3& GetMin() { return m_Min; }
			Vec3& GetMax() { return m_Max; }

			Vec3& GetRayOrigin() { return m_RayOrigin; }
			Vec3& GetRayDirection() { return m_RayDirection; }

		private:

			Vec3 m_Pos;
			Vec3 m_Color;

			Vec3 m_Min, m_Max;

			Vec3 m_RayOrigin, m_RayDirection;
			Voxel* m_Hit;
			float m_MinTime, m_MaxTime;

			float m_Size;
			bool m_Clipped;
		};

		class Ray
		{

		public:

			Ray();
			~Ray();

			void SetOwner(Voxel* a_Owner);

			Vec3& GetOrigin() { return m_Origin; }
			void SetOrigin(Vec3& a_Origin);

			Vec3& GetDirection() { return m_Direction; }
			void SetDirection(Vec3& a_Direction);

			bool CollidesWith(Voxel* a_Other);

			Voxel* GetHit();

			void Begin();

		private:

			Voxel* m_Owner;
			Vec3 m_Origin, m_Direction;
			float m_TimeMin, m_TimeMax;
			Voxel* m_Hit;
		};

	
	public:
	
		void Init();
		void Tick(float a_DT);
		void Reload();
		void PostTick();
		void Render();
		void CloseDown();

		void GenerateVoxels();

		float GetVoxelSize();

	private:

		float m_VoxelSize;
		Vec3 m_VoxelOffset[8];

		Voxel* m_Voxels;

		unsigned int m_VoxelCurrentTotal;
		unsigned int m_VoxelTotal;
		unsigned int m_VoxelDataSize;

		Ray* m_Rays;

		unsigned int m_RayTotal;

		Vec3* m_VoxelVertices;
		GLuint m_VoxelVBOVertex;

		Vec3* m_VoxelColors;
		GLuint m_VoxelVBOColor;
	
	}; // class OOPVoxels

}; // namespace Exp
	
#endif